local type = type
local pairs = pairs

setfenv(1, Dock.FunctionEnviroment)

local Core = Core

local DockClass = DockClass

Core.Version = "0.5"

local db
function Core:OnInitialize()
	db = _G.LibStub("AceDB-3.0"):New("DockDB", {
		profile = {
			SavedDocks = {},
			DockBinding1 = "",
			DockBinding2 = "",
			DockBinding3 = "",
			DockBinding4 = "",
			DockBinding5 = "",
			DockBinding6 = "",
			DockBinding7 = "",
			DockBinding8 = "",
			DockBinding9 = "",
			ModuleStates = {["*"] = true,},
			ReportFailure = true,
		},
		global = {
			Enabled = true,
			FirstRun = true,
		},
	})
	local function DatabaseCallback(event)
		--HACK: AceDB-3.0 calls callbacks before updating namespaces, thus the wrong settings are applied. This waits until the next frame to update
		self:ScheduleTimer(self.OnProfileChanged, 0, self)
	end
	db:RegisterCallback("OnProfileChanged", DatabaseCallback)
	db:RegisterCallback("OnProfileReset", DatabaseCallback)
	db:RegisterCallback("OnProfileCopied", DatabaseCallback)
	
	self.db = db
	self:SetEnabledState(db.global.Enabled)
	for name, module in self:IterateModules() do
		module:SetEnabledState(false)
		module.db = db:RegisterNamespace("Module_"..name)
	end
end


function Core:OnEnable()
	db.global.Enabled = true
	for name in pairs(db.profile.SavedDocks) do
		self:NewDock(name)
	end
	for name, module in self:IterateModules() do
		if db.profile.ModuleStates[name] then
			module:Enable()
		else
			module:Disable()
		end
	end
end

function Core:OnDisable()
	db.global.Enabled = false
	for name in self:IterateDockFrames() do
		self:RemoveDockFrame(name, true)
	end
	for name, module in self:IterateModules() do
		module:Disable()
	end
end

function Core:OnProfileChanged()
	for name in self:IterateDockFrames() do
		self:RemoveDockFrame(name, true)
	end
	for name in pairs(db.profile.SavedDocks) do
		self:NewDock(name)
	end
	for name, module in self:IterateModules() do
		module:Disable()
		if db.profile.ModuleStates[name] then
			module:Enable()
		end
	end
	_G.LibStub("AceConfigRegistry-3.0"):NotifyChange("Dock")
end

function Core:OnModuleCreated(module)
	if db then --handled in OnInitialize otherwise
		module:SetEnabledState(db.profile.ModuleStates[module.name])
		module.db = db:RegisterNamespace("Module_"..module.name)
	end
end

function Core:IsFrameDocked(frame)
	return SharedDockedFrames[frame] ~= nil
end

function Core:GetDockFromFrame(frame)
	return SharedDockedFrames[frame]
end

do
	local docks, cache = {}
	function Core:NewDock(name)
		if type(name) ~= "string" then
			error("bad argument #2 to NewDock' (string expected, got %s)", type(name))
		elseif name == "" then
			error("bad argument #2 to NewDock' (name expected, got an empty string)")
		elseif docks[name] then
			error("A dock with the name %q already exists.", _G.tostring(name))
		end
		local dock
		if not cache then
			dock = DockClass:New(name)
		else
			dock = cache[#cache]
			cache[#cache] = nil
			if #cache == 0 then
				cache = nil
			end
			dock:Initialize(name)
		end
		db.profile.SavedDocks[name] = true
		docks[name] = dock
		for parent, dockButton in pairs(DockButtons) do
			if dockButton:GetDock() == name then
				dockButton:Enable()
				if type(dockButton.OnNewDock) == "function" then
					dockButton:OnNewDock()
				end
			end
		end
		self:SendMessage("Dock_DockAdded", name)
		return dock
	end
	function Core:IterateDockFrames()
		return pairs(docks)
	end
	function Core:GetDockFrame(name)
		return docks[name]
	end
	function Core:HasDockFrame(name)
		return docks[name] ~= nil
	end
	function Core:RenameDockFrame(name, newName)
		local dock = docks[name]
		if dock and not docks[newName] then
			dock:Rename(newName)
			docks[name] = nil
			docks[newName] = dock
			db.profile.SavedDocks[name] = nil
			db.profile.SavedDocks[newName] = true
			for parent, dockButton in pairs(DockButtons) do
				if dockButton:GetDock() == name then
					dockButton:Disable()
					if type(dockButton.OnRemoveDock) == "function" then
						dockButton:OnRemoveDock()
					end
				end
				if dockButton:GetDock() == newName then
					dockButton:Enable()
					if type(dockButton.OnNewDock) == "function" then
						dockButton:OnNewDock()
					end
				end
			end
			
		end
	end
	local function CopyDockSettings(t, o)
		for k, v in pairs(t) do
			if k ~= "SavedFrames" and k ~= "SavedXOffsets" and k ~= "SavedYOffsets"
				and k ~= "SavedScales" and k ~= "XOffset" and k ~= "YOffset" then

				if type(v) == "table" then
					CopyDockSettings(v, o[k])
				else
					o[k] = v
				end
			end
		end
	end
	function Core:CopyDockFrame(name, copyName)
		local dock = docks[name]
		if dock then
			local copyDock = docks[copyName] or self:NewDock(copyName)

			CopyDockSettings(dock.db.profile, copyDock.db.profile)

			copyDock:RefreshSettings()
		end	
	end
	function Core:RemoveDockFrame(name, saveSettings)
		if docks[name] then
			local dock = docks[name]
			docks[name] = nil
			if not saveSettings then
				db.profile.SavedDocks[name] = nil
			end
			dock:Remove(saveSettings)
			for parent, dockButton in pairs(DockButtons) do
				if dockButton:GetDock() == name then
					dockButton:Disable()
					if type(dockButton.OnRemoveDock) == "function" then
						dockButton:OnRemoveDock()
					end
				end
			end
			self:SendMessage("Dock_DockRemoved", name)
			if not cache then
				cache = {}
			end
			cache[#cache + 1] = dock
		end
	end
	function Core:ExecuteDockMethod(method, ...)
		if not DockClass[method] then
			error("Docks do not have a method named %q", _G.tostring(method))
		end
		for name, dock in pairs(docks) do
			dock[method](dock, ...)
		end
	end
end


function Core:Print(first, ...)
	local string
	if first:match("%%") then
		string = first:format(...)
	else
		string = (", "):join(first, ...)
	end
	_G.DEFAULT_CHAT_FRAME:AddMessage("|cff37FDFCDock|r: "..string)
end

function Core:OnDockFail(reason, frame)
	if not db.profile.ReportFailure then
		return
	end
	local name
	if frame then
		name = frame:GetName() or "<Unnamed>"
	end
	 
	if reason == "docked" then
		self:Print("Couldn't dock frame [%s] because it's already docked.", name)
	elseif reason == "size" then
		self:Print("Couldn't dock frame [%s] because it's too large to fit. Try resizing the dock or removing already docked frames.", name)
	elseif reason == "blacklist" then
		local reason = self:GetFrameBlacklistReason(frame)
		if reason then
			self:Print("Couldn't dock frame [%s] because %s.", name, reason)
		else
			self:Print("Couldn't dock frame [%s] because it's blacklisted and not allowed to be docked.", name)
		end
	elseif reason == "protected" then
		self:Print("Couldn't dock frame [%s] because it's secure. Frames such as unitframes or action bars are secured in combat and would cause taint if docked.", name)
	elseif reason == "noframe" then
		self:Print("Couldn't find a frame to dock.")
	elseif reason == "notmovable" then
		self:Print("Couldn't dock frame [%s] because it's not movable.", name)
	elseif reason == "nodock" then
		self:Print("Couldn't dock the frame because the dock associated with this keybind doesn't exist.")
	elseif reason == "notshown" then
		self:Print("Couldn't dock frame [%s] because it's not shown.", name)
	end
end