local insert = table.insert
local remove = table.remove

local select = select
local pairs = pairs
local ipairs = ipairs
local unpack = unpack
local type = type
local pcall = pcall

local GetCursorPosition = GetCursorPosition
local GetMouseFocus = GetMouseFocus
local UIParent = UIParent
local _G = _G
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

local function safecall(...)
	local success, err = pcall(...)
	if not success then
		--do nothing
		--geterrorhandler()(err)
	end
end

setfenv(1, Dock.FunctionEnviroment)

local SharedDockedFrames = SharedDockedFrames

local Core = Core

local DockClass = DockClass

_G.LibStub("AceTimer-3.0"):Embed(DockClass)
_G.LibStub("AceHook-3.0"):Embed(DockClass)

local SetPoint


local FrameInfo = FrameInfo

local clamped = {}
local strata = {}
local centerX, centerY = {}, {}

function DockClass:GUIRemove()
	return Core:RemoveDockFrame(self.Name)
end

function DockClass:Remove(saveSettings)
	self:ReleaseAndRestoreAll(saveSettings)
	self:Hide()
	if not saveSettings then
		local current = Core.db:GetCurrentProfile()
		--perhaps an easier way to remove a namespace is in order?
		_G.rawset(Core.db.children[self.dbName].profiles, current, nil)
		_G.rawset(Core.db.sv.namespaces[self.dbName].profiles, current, nil)
		_G.rawset(Core.db.children[self.dbName], "profile", nil)
	end
	self.db = nil
end


function DockClass:Rename(name)
	local tmp = {}
	for k, v in pairs(self.db.profile) do
		tmp[k] = v
	end

	local current = Core.db:GetCurrentProfile()
	_G.rawset(Core.db.children[self.dbName].profiles, current, nil)
	_G.rawset(Core.db.sv.namespaces[self.dbName].profiles, current, nil)
	_G.rawset(Core.db.children[self.dbName], "profile", nil)

	self.Name = name
	self.dbName = "DockFrame_"..name

	self.db = Core.db.children[self.dbName] or Core.db:RegisterNamespace(self.dbName, FrameDefaults)

	local db = self.db.profile
	for k, v in pairs(tmp) do
		db[k] = v
	end
	self:UpdateTab()
end

function DockClass:Initialize(name)
	self.dbName = "DockFrame_"..name 
	self.db = Core.db.children and Core.db.children[self.dbName] or Core.db:RegisterNamespace(self.dbName, FrameDefaults)
	self.Name = name

	self.DockedFrames = {}

	self:RefreshSettings()

	self:Show()

	local delay = self.db.profile.ValidateFrameDelay
	for frameName in pairs(self.db.profile.SavedFrames) do
		self:ScheduleTimer("ValidateFrame", delay, frameName)
	end
end

do
	local function OnDragStop(self)
		self:StopMovingOrSizing()
		self.db.profile.XOffset, self.db.profile.YOffset = self:GetCenter()
		self:UpdateAnchor()
		self:UpdateAnchors()
		self.IsMovingOrResizing = nil
	end
	local function OnDragStart(self)
		self.IsMovingOrResizing = true
		self:StartMoving()
	end

	local function OnMouseDown(self, button)
		local db = self.db.profile
		local key = "Dock"..button
		if db[key] == "config" then
			AceConfigDialog:Open("Dock")
		elseif db[key] == "releaseAll" then
			self:ReleaseAndRestoreAll()
		elseif db[key] == "tabMode" then
			self:SendToTab()
		end
	end
	local function OnEnter(self)
		if not self.db.profile.Locked then
			self.DragHandle:Show()
		end
	end
	local function OnLeave(self)
		if GetMouseFocus() ~= self.DragHandle then
			self.DragHandle:Hide()
		end
	end
	function DockClass:SendToTab()
		local db = self.db.profile
		if not db.AllowTabMode then
			return
		end
		local animation = Core:GetAnimation(db.TabAnimation)
		self:SetClampedToScreen(nil)
		if animation then
			self.AnimatingRate = db.TabAnimationRate
			self.AnimationAnchor = db.VerticalAnchor..db.HorizontalAnchor
			
			local x, y = self:GetCenter()
			self:ClearAllPoints()
			self:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)

			self.EndingX, self.EndingY = db.TabX, db.TabY

			self.IsAnimating = animation
		else
			self:AcquireTab()
			self:ClearAllPoints()
			self:SetPoint("TOP", UIParent, "BOTTOM", 1000, 1000)
		end
		for i, frame in ipairs(self.DockedFrames) do
			frame:SetScale(FrameInfo.FitScale[frame])
		end
		self.ZoomsReset = true
		self.IsHidden = true
	end
	function DockClass:RestoreFromTab()
		self:ReclaimTab()
		local db = self.db.profile
		local animation = Core:GetAnimation(db.TabAnimation)
		if animation then
			self.AnimatingRate = db.TabAnimationRate
			self.AnimationAnchor = db.VerticalAnchor..db.HorizontalAnchor

			self:ClearAllPoints()
			self:SetPoint("CENTER", UIParent, "BOTTOMLEFT", db.TabX, db.TabY)

			self.EndingX, self.EndingY = db.XOffset, db.YOffset

			self.IsAnimating = animation
		else
			self:UpdateAnchor()
			self:UpdateAnchors()
		end
		self:SetClampedToScreen(true)
		self.IsHidden = nil
	end
	local function OnUpdate(self, elapsed)
		if self.IsAnimating then
			local x, y = self:GetCenter()
			local ex, ey = self.EndingX, self.EndingY

			local xDiff, yDiff = ex - x, ey - y
			if (xDiff < 0 and -xDiff or xDiff) < 10 and (yDiff < 0 and -yDiff or yDiff) < 10 then
				self.IsAnimating = nil
				if self.IsHidden then
					self:SetPoint("TOP", UIParent, "BOTTOM", 1000, 1000)
					self:AcquireTab()
				else
					self:UpdateAnchor()
					self:UpdateAnchors()
				end
			else
				self:SetPoint(
					"CENTER",
					UIParent,
					"BOTTOMLEFT",
					self.IsAnimating(x, y, ex, ey, self.AnimatingRate)
				)
			end
		elseif not self.IsHidden then
			local left, right, bottom, top = self.Left, self.Right, self.Bottom, self.Top
			local x, y = self:GetScaledCursorPosition()
			local db = self.db.profile
			local mouseDetect = db.MouseDetect

			if y > bottom and y < top and x > left and x < right then
				if not FrameInfo.MovingFrame then
					if self.lastMouseFocus then
						self:OnReceiveFrame(self.lastMouseFocus, x, y)
						self.lastMouseFocus = nil
					end
					self:HideInsert()
				elseif FrameInfo.MovingFrame ~= self then
					self.lastMouseFocus = FrameInfo.MovingFrame
					self:ShowInsert(self.lastMouseFocus, x, y, FrameInfo.LastReason ~= nil)
				end
				self:SetBackdropColor(unpack(db.BackdropHighlightColor))
				self:SetBackdropBorderColor(unpack(db.BackdropBorderHighlightColor))
				
			elseif y < bottom - mouseDetect or y > top + mouseDetect or x < left - mouseDetect or x > right + mouseDetect then
				if not self.ZoomsReset then
					for i, frame in ipairs(self.DockedFrames) do
						frame:SetScale(self:GetZoomLocked(frame) or FrameInfo.FitScale[frame])
					end
					self.ZoomsReset = true
				end
				if db.AutoTab then
					self.TabTime = self.TabTime - elapsed
					if self.TabTime < 0 then
						self.TabTime = db.AutoTabDelay
						self:SendToTab()
					end
				end
				return
			else
				self.lastMouseFocus = nil
				self:SetBackdropColor(unpack(db.BackdropColor))
				self:SetBackdropBorderColor(unpack(db.BackdropBorderColor))
				self:HideInsert()
			end
			self.TabTime = db.AutoTabDelay
			local dockedFrames = self.DockedFrames
			if #dockedFrames > 0 then
				if self.IsMovingOrResizing then
					for i, frame in ipairs(dockedFrames) do
						frame:SetScale(FrameInfo.FitScale[frame])
					end
				else
					local mouseOverScale = 150 * (1 - db.MouseOverSize)
					for i, frame in ipairs(dockedFrames) do
						if frame:GetNumPoints() > 1 then
							self:UpdateAnchors()
						end
						local zoomLocked = self:GetZoomLocked(frame)
						if zoomLocked then
							frame:SetScale(zoomLocked)
						else
							local cx, cy = centerX[frame], centerY[frame]
							local xDiff, yDiff = cx - x, cy - y
							xDiff = (xDiff < 0 and -xDiff or xDiff) --twice as fast as abs
							yDiff = (yDiff < 0 and -yDiff or yDiff)
							local diff = xDiff > yDiff and xDiff or yDiff
							
							local fitScale = FrameInfo.FitScale[frame]
							local newScale = ((mouseDetect - diff) / mouseOverScale) * fitScale

							frame:SetScale(newScale > fitScale and newScale or fitScale)
						end
					end
				end
				self.ZoomsReset = false
			end
		end
	end
	local function DragOnMouseDown(self)
		local parent = self:GetParent()
		parent.IsMovingOrResizing = true

		parent:StartSizing("BOTTOMRIGHT")
	end
	local function DragOnMouseUp(self)
		local parent = self:GetParent()
		parent:StopMovingOrSizing()

		parent.IsMovingOrResizing = false

		local db = parent.db.profile

		db.XOffset, db.YOffset = parent:GetCenter()

		db.Height = parent:GetHeight()
		db.Width = parent:GetWidth()

		parent:RefreshSettings()
	end
	function DockClass:New(name)
		if self ~= DockClass then
			error("New can only be called on the DockClass")
		end
		
		local dock = CreateFrame("Frame", nil, UIParent)

		dock:SetMinResize(50, 50)

		Core:AddFrameToBlacklist(dock, "docks can't be docked")

		dock:SetClampedToScreen(true)
		
		dock:RegisterForDrag("LeftButton")
		dock:SetFrameStrata("BACKGROUND")

		dock:SetScript("OnDragStart", OnDragStart)
		dock:SetScript("OnDragStop", OnDragStop)
		dock:SetScript("OnMouseDown", OnMouseDown)
		
		dock:SetScript("OnUpdate", OnUpdate)
		dock:SetScript("OnLeave", OnLeave)
		dock:SetScript("OnEnter", OnEnter)

		local drag = CreateFrame("Frame", nil, dock)
		drag:Hide()
		drag:SetHeight(16)
		drag:SetWidth(16)
		drag:EnableMouse(true)
		drag:SetPoint("BOTTOMRIGHT", dock, "BOTTOMRIGHT")

		drag:SetScript("OnMouseDown", DragOnMouseDown)
		drag:SetScript("OnMouseUp", DragOnMouseUp)
		drag:SetScript("OnLeave", drag.Hide)

		local texture = drag:CreateTexture(nil, "BACKGROUND")
		texture:SetBlendMode("ADD")
		texture:SetAllPoints(drag)
		texture:SetTexture[[Interface\AddOns\Dock\Images\DragHandle]]

		dock.DragHandle = drag

		local anchor = CreateFrame("Frame", nil, dock)
		anchor:SetWidth(1)
		anchor:SetHeight(1)
		dock.Anchor = anchor

		for k, v in pairs(self) do
			dock[k] = v
		end


		dock:Initialize(name)
		return dock
	end
end

do
	local function Top(self, frame)
		local dockedFrames = self.DockedFrames
		if #dockedFrames > 0 then
			local lastFrame = dockedFrames[#dockedFrames]
			local fscale = frame:GetScale()
			return self:GetCenter() * self:GetScale(),
				lastFrame:GetBottom() * lastFrame:GetScale() - (frame:GetHeight() / 2) * fscale - self.db.profile.Border * fscale
		end
		local scale, fscale = self:GetScale(), frame:GetScale()
		return self:GetCenter() * scale,
			self:GetTop() * scale - (frame:GetHeight() / 2) * fscale - self.db.profile.Border * fscale
	end
	local function Bottom(self, frame)
		local dockedFrames = self.DockedFrames
		if #dockedFrames > 0 then
			local lastFrame = dockedFrames[#dockedFrames]
			local fscale = frame:GetScale()
			return self:GetCenter() * self:GetScale(),
				lastFrame:GetTop() * lastFrame:GetScale() + (frame:GetHeight() / 2) * fscale + self.db.profile.Border * fscale
		end
		local scale, fscale = self:GetScale(), frame:GetScale()
		return self:GetCenter() * scale, self:GetBottom() * scale + (frame:GetHeight() / 2) * fscale + self.db.profile.Border * fscale
	end
	local function Left(self, frame)
		local dockedFrames = self.DockedFrames
		if #dockedFrames > 0 then
			local lastFrame = dockedFrames[#dockedFrames]
			local fscale = frame:GetScale()
			return lastFrame:GetRight() * lastFrame:GetScale() + (frame:GetWidth() / 2) * fscale + self.db.profile.Border * fscale,
				select(2, self:GetCenter()) * self:GetScale()
		end
		local scale, fscale = self:GetScale(), frame:GetScale()
		return self:GetLeft() * scale + (frame:GetWidth() / 2) * fscale + self.db.profile.Border * fscale, 
			select(2, self:GetCenter()) * scale
	end
	local function Right(self, frame)
		local dockedFrames = self.DockedFrames
		if #dockedFrames > 0 then
			local lastFrame = dockedFrames[#dockedFrames]
			local fscale = frame:GetScale()
			return lastFrame:GetLeft() * lastFrame:GetScale() - (frame:GetWidth() / 2) * fscale - self.db.profile.Border * fscale,
				select(2, self:GetCenter()) * self:GetScale()
		end
		local scale, fscale = self:GetScale(), frame:GetScale()
		return self:GetRight() * scale - (frame:GetWidth() / 2) * fscale - self.db.profile.Border * fscale, 
			select(2, self:GetCenter()) * scale
	end
	function DockClass:RefreshSettings()
		local db = self.db.profile

		self:SetWidth(db.Width)
		self:SetHeight(db.Height)

		do
			local backdrop = db.Backdrop
			local bg, edge = backdrop.bgFile, backdrop.edgeFile

			backdrop.bgFile, backdrop.edgeFile = MediaHandler:Fetch("background", bg), MediaHandler:Fetch("border", edge)
			self:SetBackdrop(backdrop)

			backdrop.bgFile, backdrop.edgeFile = bg, edge

			self:SetClampRectInsets(backdrop.insets.left, -backdrop.insets.right, -backdrop.insets.top, backdrop.insets.bottom)
		end

		self:SetBackdropColor(unpack(db.BackdropColor))
		self:SetBackdropBorderColor(unpack(db.BackdropBorderColor))

		self:SetMovable(not db.Locked)
		self:EnableMouse(not db.Locked)
		self:SetResizable(not db.Locked)

		if db.Orientation == "Vertical" then
			self.GetAnimationDockingPoints  = db.VerticalAnchor == "TOP" and Top or Bottom
		else
			self.GetAnimationDockingPoints  = db.HorizontalAnchor == "LEFT" and Left or Right
		end
		
		self.TabTime = db.AutoTabDelay

		self:UpdateAnchor()

		self:UpdateCurrentDocks()

		self:UpdateAnchors()

		self:UpdateTab()

		self:ReleaseExtraFrames()
	end
end

function DockClass:GetScaledCursorPosition()
	local x, y = GetCursorPosition()
	local scale = self:GetEffectiveScale()
	return x / scale, y / scale
end

do
	local oppositePoint = {
		LEFT = "RIGHT",
		RIGHT = "LEFT",
		TOP = "BOTTOM",
		BOTTOM = "TOP",
	}
	local InsertFrame = CreateFrame("Frame")
	InsertFrame:SetFrameStrata("TOOLTIP")
	InsertFrame:SetWidth(35)
	InsertFrame:SetHeight(35)

	local InsertArrow = InsertFrame:CreateTexture()
	InsertArrow:SetTexture[[Interface\Addons\Dock\Images\Arrow.tga]]
	InsertArrow:SetAllPoints(InsertFrame)

	local NoFit = InsertFrame:CreateTexture(nil, "OVERLAY")
	NoFit:SetTexture[[Interface\Addons\Dock\Images\NoFit.tga]]
	NoFit:SetAllPoints(InsertArrow)
	NoFit:SetVertexColor(1, 0, 0)

	SetPoint = InsertFrame.SetPoint
	local lastShown
	function DockClass:ShowInsert(frame, x, y, forceNoFit)
		if not forceNoFit and self:WillFit(frame) then
			NoFit:Hide()
		else
			NoFit:Show()
		end
		InsertFrame:ClearAllPoints()
		local dockedFrames = self.DockedFrames
		local db = self.db.profile
		local tab = self.Tab
		local vAnchor = db.VerticalAnchor
		local hAnchor = db.HorizontalAnchor
		local orientation = tab and db.TabOriention or db.Orientation
		if orientation == "Vertical" then
			if hAnchor == "RIGHT" then
				InsertArrow:SetTexCoord(0, 0, 0, 1, 1, 0, 1, 1)
			else
				InsertArrow:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
			end
			if tab then
				InsertFrame:SetPoint(hAnchor, tab, oppositePoint[hAnchor])
			else
				if #dockedFrames > 0 then
					local pos = self:GetInsertPosition(x, y)
					if pos > #dockedFrames then
						local frame = dockedFrames[pos - 1]
						InsertFrame:SetPoint(hAnchor, frame, oppositePoint[vAnchor]..oppositePoint[hAnchor])
					else
						local frame = dockedFrames[pos]
						InsertFrame:SetPoint(hAnchor, frame, vAnchor..oppositePoint[hAnchor])
					end
				else
					InsertFrame:SetPoint(hAnchor, self, vAnchor..oppositePoint[hAnchor])
				end
			end
			
		else
			if vAnchor == "TOP" then
				InsertArrow:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)
			else
				InsertArrow:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
			end
			if tab then
				InsertFrame:SetPoint(vAnchor, tab, oppositePoint[vAnchor])
			else
				if #dockedFrames > 0 then
					local pos = self:GetInsertPosition(x, y)
					if pos > #dockedFrames then
						local frame = dockedFrames[pos - 1]
						InsertFrame:SetPoint(vAnchor, frame, oppositePoint[vAnchor]..oppositePoint[hAnchor])
					else
						local frame = dockedFrames[pos]
						InsertFrame:SetPoint(vAnchor, frame, oppositePoint[vAnchor]..hAnchor)
					end
				else
					InsertFrame:SetPoint(vAnchor, self, oppositePoint[vAnchor]..hAnchor)
				end
			end
		end
		InsertArrow:SetVertexColor(unpack(db.ArrowColor))
		InsertFrame:Show()
		lastShown = self
	end
	function DockClass:HideInsert()
		if lastShown == self then
			InsertFrame:Hide()
		end
	end
end

do
	local lockedZooms = {}
	function DockClass:GetZoomLocked(frame)
		return lockedZooms[frame]
	end
	function DockClass:LockZoom(frame, scale)
		lockedZooms[frame] = scale or frame:GetScale()
		self:UpdateAnchors()
	end
	function DockClass:UnlockZoom(frame)
		lockedZooms[frame] = nil
		self:UpdateAnchors()
	end
end

function DockClass:GetInsertPosition(x, y)
	local dockedFrames = self.DockedFrames
	local db = self.db.profile
	local coord, less, num
	if db.Orientation == "Vertical" then
		coord = y
		num = 2
		less = db.VerticalAnchor == "TOP"
	else
		coord = x
		num = 1
		less = db.HorizontalAnchor == "RIGHT"
	end
	
	if #dockedFrames > 0 then
		for i=#dockedFrames, 1, -1 do
			local frame = dockedFrames[i]
			if less then
				if coord < select(num, frame:GetCenter()) * frame:GetScale() then
					return i + 1
				end
			else
				if coord > select(num, frame:GetCenter()) * frame:GetScale() then
					return i + 1
				end
			end
		end
	end
	return 1
end


function DockClass:ReleaseAndRestoreAll(saveSettings)
	local dockedFrames = self.DockedFrames
	for i=#dockedFrames, 1, -1 do
		self:ReleaseAndRestore(dockedFrames[i], saveSettings)
	end
end

function DockClass:ReleaseExtraFrames()
	local dockedFrames = self.DockedFrames
	local db = self.db.profile
	local method, maxSize
	local border = db.Border
	if db.Orientation == "Vertical" then
		method = "GetHeight"
		maxSize = db.Height - border
		
	else
		method = "GetWidth"
		maxSize = db.Width - border
	end
	local size = -border
	for i, frame in ipairs(dockedFrames) do
		size = size + (border + (frame[method](frame) * FrameInfo.FitScale[frame]))
	end
	if size > maxSize then
		for i=#dockedFrames, 1, -1 do
			local frame = dockedFrames[i]
			size = size - (border + (frame[method](frame) * FrameInfo.FitScale[frame]))
			self:ReleaseAndRestore(frame)
			if size <= maxSize then
				return
			end
		end
	end
end

function DockClass:WillFit(frame)
	local dockedFrames = self.DockedFrames
	local db = self.db.profile
	local method, maxSize, size
	local border = db.Border
	if db.Orientation == "Vertical" then
		method = "GetHeight"
		maxSize = db.Height - border
		size = frame:GetHeight() * self:CalculateFitScale(frame)
		
	else
		method = "GetWidth"
		maxSize = db.Width - border
		size = frame:GetWidth() * self:CalculateFitScale(frame)
	end
	for i, f in ipairs(dockedFrames) do
		size = size + (border + (f[method](f) * FrameInfo.FitScale[f]))
		if size > maxSize then
			return false
		end
	end
	return size < maxSize
end

function DockClass:OnReceiveFrame(frame, x, y)
	if FrameInfo.LastReason then
		Core:OnDockFail(FrameInfo.LastReason, frame)
		FrameInfo.LastReason = nil
	else
		self:DockFrame(frame, x and y and self:GetInsertPosition(x, y) or 1, FrameInfo.MovingX, FrameInfo.MovingY)
	end
end

function DockClass:IsFrameDocked(frame)
	return SharedDockedFrames[frame] == self
end

do
	local NoFit
	local lastTimer
	function DockClass:SendToDock(frame, noAnimation, button)
		if self:IsFrameDocked(frame) then
			Core:OnDockFail("docked", frame)
			return
		end
		if not self:WillFit(frame) then
			if button then
				if not NoFit then
					NoFit = button:CreateTexture(nil, "OVERLAY")
					NoFit:SetTexture([[Interface\Addons\Dock\Images\NoFit.tga]])
					NoFit:SetVertexColor(1, 0, 0)
				end
				NoFit:SetAllPoints(button)
				NoFit:Show()
				if lastTimer then
					self:CancelTimer(lastTimer, true)
				end
				lastTimer = self:ScheduleTimer(NoFit.Hide, 1, NoFit)
			end
			Core:OnDockFail("size", frame)
			return
		end
		if SharedDockedFrames[frame] then
			local dock = SharedDockedFrames[frame]
			dock:ReleaseFrame(frame)
			dock:RestorePosition(frame, true)
			if SharedDockedFrames[frame] then
				return --shouldn't happen
			end
		end
		if noAnimation or self.db.profile.DockAnimation == "None" or self.Tab then
			local scale, xoff, yoff
			if AnimatingRestorations[frame] or AnimatingDocks[frame] then
				AnimatingRestorations[frame] = nil
				AnimatingDocks[frame] = nil
				scale = FrameInfo.Scale[frame]
				xoff, yoff = FrameInfo.XOffset[frame], FrameInfo.YOffset[frame]
			end
			local success = self:DockFrame(frame, nil, frame:GetCenter())
			if scale then
				FrameInfo.Scale[frame] = scale
			end
			if xoff then
				FrameInfo.XOffset[frame], FrameInfo.YOffset[frame] = xoff, yoff
			end
			if success then
				Core:SendMessage("Dock_FrameSentToDock", frame, self, button)
			end
			return success
		end
		self:AnimateFrame(frame)
		Core:SendMessage("Dock_FrameSentToDock", frame, self, button)
		return true
	end
end

function DockClass:VerifyDockedFrameHide(frame)
	if not frame:IsShown() then
		self:ReleaseFrame(frame)
		self:RestorePosition(frame, true)
	end
end

function DockClass:OnDockedFrameHide(frame)
	if not frame:IsShown() then
		self:ScheduleTimer("VerifyDockedFrameHide", 0, frame)
	end
end

function DockClass:UpdateCurrentDocks()
	for i, frame in ipairs(self.DockedFrames) do
		local neededScale = self:CalculateFitScale(frame)
		frame:SetScale(neededScale)
		FrameInfo.FitScale[frame] = neededScale
	end
end

function DockClass:UpdateAnchor(ignoreTab)
	self:ClearAllPoints()
	local db = self.db.profile
	if not db.XOffset then
		db.XOffset = _G.GetScreenWidth() / 2
	end
	if not db.YOffset then
		db.YOffset = _G.GetScreenHeight() / 2
	end
	if not db.TabX then
		db.TabX = db.XOffset
	end
	if not db.TabY then
		db.TabY = db.YOffset
	end
	if not ignoreTab and self.Tab then
		self:SetPoint("TOP", UIParent, "BOTTOM", 1000, 1000)
	else
		self:SetPoint(
			db.VerticalAnchor..db.HorizontalAnchor,
			UIParent, "BOTTOMLEFT",
			db.XOffset + (db.HorizontalAnchor == "RIGHT" and self:GetWidth() or -self:GetWidth()) / 2,
			db.YOffset + (db.VerticalAnchor == "TOP" and self:GetHeight() or -self:GetHeight()) / 2
		)
		self.Left, self.Right, self.Bottom, self.Top = self:GetLeft(), self:GetRight(), self:GetBottom(), self:GetTop()
	end
end


function DockClass:UpdateAnchors()
	local dockedFrames = self.DockedFrames
	local db = self.db.profile
	if self.Tab then
		self:UpdateAnchor(true)
	end
	
	local point = db.VerticalAnchor..db.HorizontalAnchor
	local rpoint, yOffset, xOffset
	local anchorXOffset
	if db.Orientation == "Vertical" then
		rpoint = (db.VerticalAnchor == "TOP" and "BOTTOM" or "TOP")..db.HorizontalAnchor
		yOffset = db.VerticalAnchor == "TOP" and -db.Border or db.Border
		xOffset = 0
	else
		rpoint = db.VerticalAnchor..(db.HorizontalAnchor == "LEFT" and "RIGHT" or "LEFT")
		xOffset = db.HorizontalAnchor == "RIGHT" and -db.Border or db.Border
		yOffset = 0
	end

	self.Anchor:ClearAllPoints()
	self.Anchor:SetPoint(point, self, point, 
		db.HorizontalAnchor == "RIGHT" and -db.Border / 2 or db.Border / 2,
		db.VerticalAnchor == "TOP" and -db.Border / 2 or db.Border / 2
	)
	for i, frame in ipairs(dockedFrames) do
		frame:ClearAllPoints()
		if i == 1 then
			SetPoint(frame, point, self.Anchor, point, 0, 0)
		else
			SetPoint(frame, point, dockedFrames[i - 1], rpoint, xOffset, yOffset)
		end
		local scale = self:GetZoomLocked(frame) or FrameInfo.FitScale[frame]
		frame:SetScale(scale)
		local x, y = frame:GetCenter()
		centerX[frame], centerY[frame] = x * scale, y * scale
	end
	self.ZoomsReset = false
	if self.Tab then
		self:UpdateAnchor()
	end
end

function DockClass:ReleaseAndRestore(frame, saveSettings)
	self:ReleaseFrame(frame, saveSettings)
	self:RestorePosition(frame)
end

function DockClass:CalculateFitScale(frame)
	local db = self.db.profile
	local width
	if db.ForceWidth then
		width = db.ForcedWidth / frame:GetWidth()
	end
	if db.ForceHeight then
		local height = db.ForcedHeight / frame:GetHeight()
		return (not width or width < height) and height or width
	end
	if width then
		return width
	end
	return ((db.Orientation == "Vertical" and db.Width or db.Height) - db.Border) / (db.Orientation == "Vertical" and frame:GetWidth() or frame:GetHeight())
end

function DockClass:DockFrame(frame, pos, x, y)
	if self:IsFrameDocked(frame) then
		Core:OnDockFail("docked", frame)
		return false
	end
	if not self:WillFit(frame) then
		Core:OnDockFail("size", frame)
		return false
	end
	if SharedDockedFrames[frame] then
		local dock = SharedDockedFrames[frame]
		dock:ReleaseFrame(frame)
		dock:RestorePosition(frame, true)
		if SharedDockedFrames[frame] then
			return --shouldn't happen
		end
	end
	local scale = frame:GetScale()
	FrameInfo.Scale[frame] = scale
	FrameInfo.XOffset[frame], FrameInfo.YOffset[frame] = (x or FrameInfo.XOffset[frame] or 0) / scale, (y or FrameInfo.YOffset[frame] or 0) / scale
	local db = self.db.profile
	

	local neededScale = self:CalculateFitScale(frame)
	frame:SetScale(neededScale)
	FrameInfo.FitScale[frame] = neededScale

	clamped[frame] = frame:IsClampedToScreen()
	strata[frame] = frame:GetFrameStrata()
	frame:SetClampedToScreen(nil)
	frame:SetFrameStrata("BACKGROUND")

	self:AcquireButtonOverlay(frame)

	local dockedFrames = self.DockedFrames
	if not pos or pos > #dockedFrames + 1 then
		pos = #dockedFrames + 1
	end
	insert(dockedFrames, pos, frame)

	self:UpdateAnchors()
	local name = frame:GetName()
	if name then
		local db = self.db.profile
		db.SavedFrames[name] = pos
		db.SavedXOffsets[name] = FrameInfo.XOffset[frame]
		db.SavedYOffsets[name] = FrameInfo.YOffset[frame]
		db.SavedScales[name] = FrameInfo.Scale[frame]
	end
	self:HookScript(frame, "OnHide", "OnDockedFrameHide")

	self:RawHook(frame, "SetPoint", "OnDockedFrameSetPoint", true)

	SharedDockedFrames[frame] = self
	Core:SendMessage("Dock_FrameDocked", frame, self)
	return true
end

function DockClass:OnDockedFrameSetPoint(frame, ...)
	self.hooks[frame].SetPoint(frame, ...)
	self:UpdateAnchors()
end

function DockClass:ReleaseFrame(frame)
	local dockedFrames = self.DockedFrames
	local found
	for i, f in ipairs(dockedFrames) do
		if frame == f then
			remove(dockedFrames, i)
			found = true
			break
		end
	end
	if not found then
		return false
	end
	self:UpdateAnchors()
	self:ReclaimButtonOverlay(self:GetActiveOverlay(frame))
	local name = frame:GetName()
	if name and not saveSettings then
		local db = self.db.profile
		db.SavedFrames[name] = nil
		db.SavedXOffsets[name] = nil
		db.SavedYOffsets[name] = nil
		db.SavedScales[name] = nil
	end
	self:Unhook(frame, "OnHide")
	self:Unhook(frame, "SetPoint")

	self:UnlockZoom(frame)

	frame:SetClampedToScreen(clamped[frame])
	frame:SetFrameStrata(strata[frame])

	SharedDockedFrames[frame] = nil
	Core:SendMessage("Dock_FrameReleased", frame, self)
	return true
end

do
	local AnimationFrame = CreateFrame("Frame")
	AnimationFrame:Hide()

	local restores, rates = AnimatingRestorations, {}
	local sendToDock, frameToDock = AnimatingDocks, {}

	AnimationFrame:SetScript("OnUpdate", function(self)
		local hide = true
		for frame, animation in pairs(restores) do
			hide = false
			local x, y = frame:GetCenter()
			local origScale = FrameInfo.Scale[frame]

			local ex, ey = FrameInfo.XOffset[frame] * origScale, FrameInfo.YOffset[frame] * origScale
			local xDiff, yDiff = ex - x, ey - y
			if (xDiff < 0 and -xDiff or xDiff) < 10 and (yDiff < 0 and -yDiff or yDiff) < 10 then
				restores[frame] = nil
				frame:SetPoint("CENTER", UIParent, "BOTTOMLEFT", ex, ey)
				frame:SetScale(FrameInfo.Scale[frame])
				local onStop = frame:GetScript("OnDragStop") or frame:GetScript("OnMouseUp")
				if onStop then
					FrameInfo.IgnoreHook = true
					_G.this = frame
					_G.arg1 = "LeftButton"
					safecall(onStop, frame, "LeftButton")
					FrameInfo.IgnoreHook = false
				end
				Core:SendMessage("Dock_FrameRestored", frame)
			else
				local scale = frame:GetScale()
				local newScale = scale + (origScale - scale) * rates[frame]
				frame:SetScale(newScale)
				local ax, ay = animation((x * scale) / newScale, (y * scale) / newScale, ex, ey, rates[frame])
				frame:SetPoint(
					"CENTER",
					UIParent,
					"BOTTOMLEFT",
					ax, ay
				)
			end
		end
		for frame, animation in pairs(sendToDock) do
			hide = false
			local scale = frame:GetScale()
			local x, y = frame:GetCenter()
			x, y = x * scale, y * scale
			
			local dock = frameToDock[frame]
			local ex, ey = dock:GetAnimationDockingPoints(frame)

			local xDiff, yDiff = ex - x, ey - y
			if (xDiff < 0 and -xDiff or xDiff) < 10 and (yDiff < 0 and -yDiff or yDiff) < 10 then
				sendToDock[frame] = nil
				frame:SetScale(FrameInfo.Scale[frame])
				if not dock:DockFrame(frame) then
					frame:SetScale(FrameInfo.FitScale[frame])
					dock:RestorePosition(frame)
				end
			else
				local newScale = scale + (FrameInfo.FitScale[frame] - scale) * rates[frame]
				frame:SetScale(newScale)
				local ax, ay = animation(x, y, ex, ey, rates[frame])
				frame:SetPoint(
					"CENTER",
					UIParent,
					"BOTTOMLEFT",
					ax / newScale, ay / newScale
				)
			end
		end
		if hide then self:Hide() end
	end)
	function DockClass:IsFrameAnimating(frame)
		return sendToDock[frame] ~= nil or restores[frame] ~= nil
	end
	function DockClass:StopFrameAnimation(frame)
		if frame then
			sendToDock[frame] = nil
			restores[frame] = nil
		end
	end
	function DockClass:RestorePosition(frame, noAnimation)
		local animation = Core:GetAnimation(self.db.profile.Animation)
		if not noAnimation and animation and not self.Tab then
			restores[frame] = animation
			rates[frame] = self.db.profile.AnimationRate
			local x, y = frame:GetCenter()
			frame:ClearAllPoints()
			frame:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)
			sendToDock[frame] = nil
			AnimationFrame:Show()
		else
			frame:ClearAllPoints()
			local scale = FrameInfo.Scale[frame]
			frame:SetPoint("CENTER", UIParent, "BOTTOMLEFT", FrameInfo.XOffset[frame] * scale, FrameInfo.YOffset[frame] * scale)
			frame:SetScale(scale)
			local onStop = frame:GetScript("OnDragStop") or frame:GetScript("OnMouseUp")
			if onStop then
				FrameInfo.IgnoreHook = true
				_G.this = frame
				_G.arg1 = "LeftButton"
				safecall(onStop, frame, "LeftButton")
				FrameInfo.IgnoreHook = false
			end
			Core:SendMessage("Dock_FrameRestored", frame)
		end
	end
	function DockClass:AnimateFrame(frame)
		local db = self.db.profile
		local animation = Core:GetAnimation(db.DockAnimation)
		local x, y = frame:GetCenter()
		if not sendToDock[frame] then
			FrameInfo.XOffset[frame], FrameInfo.YOffset[frame] = x, y
			FrameInfo.Scale[frame] = frame:GetScale()
			FrameInfo.FitScale[frame] = ((db.Orientation == "Vertical" and db.Width or db.Height) - db.Border) / (db.Orientation == "Vertical" and frame:GetWidth() or frame:GetHeight())
		end
		sendToDock[frame] = animation
		rates[frame] = db.DockAnimationRate
		frameToDock[frame] = self
		restores[frame] = nil

		frame:ClearAllPoints()
		frame:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)

		AnimationFrame:Show()
	end
end

function DockClass:ValidateFrame(frameName)
	local frame = _G[frameName]
	if frame and frame:GetNumPoints() > 0 and frame:IsShown() and not self:IsFrameDocked(frame) then
		local db = self.db.profile
		FrameInfo.Scale[frame] = db.SavedScales[frameName]
		self:DockFrame(frame, db.SavedFrames[frameName], db.SavedXOffsets[frameName], db.SavedYOffsets[frameName])
	end
end

function DockClass:OnCreateFrame(objectType, frameName, ...)
	if self.db.profile.SavedFrames[frameName] then
		self:ScheduleTimer("ValidateFrame", self.db.profile.ValidateFrameDelay, frameName)
	end
end


do
	local function OnClick(self, button)
		local parent = self:GetParent()
		local db = parent.db.profile
		local key = "Overlay"..button
		if db[key] == "release" then
			parent:ReleaseAndRestore(self.Frame)
		elseif db[key] == "lockZoom" then
			if parent:GetZoomLocked(self.Frame) then
				parent:UnlockZoom(self.Frame)
			else
				parent:LockZoom(self.Frame)
			end
		elseif db[key] == "releaseAll" then
			parent:ReleaseAndRestoreAll()
		elseif db[key] == "tabMode" then
			parent:SendToTab()
		end
		
	end
	local function OnEnter(self)
		self:SetBackdropBorderColor(unpack(self:GetParent().db.profile.OverlayHighlightColor))
	end
	local function OnLeave(self)
		self:SetBackdropBorderColor(unpack(self:GetParent().db.profile.OverlayColor))
	end
	local function OnDragStart(self)
		local frame = self.Frame
		if frame:IsMovable() then
			local start = frame:GetScript("OnDragStart") or frame:GetScript("OnMouseDown") or frame.StartMoving
			if start then
				self.Dragging = frame
				self:GetParent():ReleaseFrame(frame)
				frame:SetScale(FrameInfo.Scale[frame])
				_G.this = frame --sadly some frames don't pass the self argument correctly and still use "this", even after 11 months of tbc; omen - i'm looking at you
				_G.arg1 = "LeftButton"
				safecall(start, frame, "LeftButton")
			end
		end
		
	end
	local function OnDragStop(self)
		if self.Dragging then
			local frame = self.Dragging
			local stop = frame:GetScript("OnDragStop") or frame:GetScript("OnMouseUp") or frame.StopMovingOrSizing
			if stop then
				_G.this = frame
				_G.arg1 = "LeftButton"
				safecall(stop, frame, "LeftButton") 
			end
			self.Dragging = nil
			self:GetParent():ReclaimButtonOverlay(self)
		end
	end
	local backdrop = {edgeFile = [[Interface/Tooltips/UI-Tooltip-Border]], edgeSize = 10, insets={left=-5, right=-5, top=-5, bottom=-5}}
	local function NewOverlay(frame)
		local b = CreateFrame("Button")
		b:SetBackdrop(backdrop)
		b:SetBackdropBorderColor(0, 0, 0, 0)
		b:SetScript("OnClick", OnClick)
		b:SetScript("OnEnter", OnEnter)
		b:SetScript("OnLeave", OnLeave)
		b:SetScript("OnDragStart", OnDragStart)
		b:SetScript("OnDragStop", OnDragStop)
		b:RegisterForDrag("LeftButton")
		b:RegisterForClicks("LeftButtonUp", "MiddleButtonUp", "RightButtonUp")
		return b
	end
	local overlays, activeOverlays = {}, {}
	function DockClass:AcquireButtonOverlay(frame)
		local overlay
		if #overlays > 0 then
			overlay = overlays[#overlays]
			overlays[#overlays] = nil
		else
			overlay = NewOverlay()
		end
		overlay:SetParent(self)
		overlay:ClearAllPoints()
		overlay:SetPoint("TOPLEFT", frame, "TOPLEFT", -2, 2)
		overlay:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 2, -2)
		overlay:SetFrameStrata("LOW")

		overlay:SetBackdropBorderColor(unpack(self.db.profile.OverlayColor))
		overlay:Show()
		overlay.Frame = frame
		activeOverlays[frame] = overlay

		
		return overlay
	end
	function DockClass:GetActiveOverlay(frame)
		return activeOverlays[frame]
	end
	function DockClass:ReclaimButtonOverlay(overlay)
		if not overlay.Dragging then
			overlay:Hide()
			overlay:ClearAllPoints()
			overlays[#overlays + 1] = overlay
			activeOverlays[overlay.Frame] = nil
		end
	end
end

do
	local tabs, activeTabs = {}, {}
	local tabNum = 0
	local function OnClick(self, button)
		if button == "RightButton" then
			AceConfigDialog:Open("Dock")
		else
			self.Dock:RestoreFromTab()
		end
		
		
	end
	local function OnDragStop(self)
		self:StopMovingOrSizing()
		local db = self.Dock.db.profile
		db.TabX, db.TabY = self:GetCenter()
	end
	local function OnUpdate(self)
		local dock = self.Dock
		local left, right, bottom, top = self:GetLeft(), self:GetRight(), self:GetBottom(), self:GetTop()
		local x, y = dock.GetScaledCursorPosition(self)
		local db = dock.db.profile

		if y > bottom and y < top and x > left and x < right then
			if not FrameInfo.MovingFrame then
				if dock.lastMouseFocus then
					dock:OnReceiveFrame(dock.lastMouseFocus)
					dock.lastMouseFocus = nil
				end
				dock:HideInsert()
			elseif FrameInfo.MovingFrame ~= dock and FrameInfo.MovingFrame ~= self then
				dock.lastMouseFocus = FrameInfo.MovingFrame
				dock:ShowInsert(dock.lastMouseFocus, x, y, FrameInfo.LastReason ~= nil)
			end
			self:SetBackdropColor(unpack(db.TabHighlightColor))
			self:SetBackdropBorderColor(unpack(db.TabBorderHighlightColor))
		else
			dock.lastMouseFocus = nil
			self:SetBackdropColor(unpack(db.TabColor))
			self:SetBackdropBorderColor(unpack(db.TabBorderColor))
			dock:HideInsert()
		end
	end
	local function NewTab()
		tabNum = tabNum + 1
		local tab = CreateFrame("Button", "DockTab"..tabNum, UIParent)
		tab:SetMovable(true)
		tab:SetNormalTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		tab:SetClampedToScreen(true)
		do
			local t = tab:GetNormalTexture()
			t:SetVertexColor(.1, .1, .1)
			t:ClearAllPoints()
			t:SetPoint("TOPLEFT", 2, -2)
			t:SetPoint("BOTTOMRIGHT", -2, 2)
		end
		
		tab:SetScript("OnClick", OnClick)
		tab:SetScript("OnDragStart", tab.StartMoving)
		tab:SetScript("OnDragStop", OnDragStop)
		tab:SetScript("OnUpdate", OnUpdate)
		tab:RegisterForDrag("LeftButton")
		tab:RegisterForClicks("LeftButtonUp", "MiddleButtonUp", "RightButtonUp")
		
		return tab
	end
	function DockClass:UpdateTab()
		local tab = self.Tab
		if not tab then
			return
		end
		local db = self.db.profile
		if not db.AllowTabMode then
			return self:RestoreFromTab()
		end
		tab:SetFont(MediaHandler:Fetch("font", db.TabFont), db.TabFontSize)
		if db.TabOrientation == "Vertical" then
			local label = "" --not called often enough to worry about garbage
			for letter in self.Name:gmatch(".") do
				label = label .. letter .. "\n"
			end
			tab:SetText(label)
			local fs = tab:GetFontString()
			tab:SetHeight(fs:GetHeight() + db.TabSize)
			tab:SetWidth(fs:GetWidth() + db.TabSize)
		else
			tab:SetText(self.Name)
			tab:SetHeight(30)
			local fs = tab:GetFontString()
			tab:SetWidth(fs:GetWidth() + db.TabSize)
			tab:SetHeight(fs:GetHeight() + db.TabSize)
		end
		tab:SetTextColor(unpack(db.TabFontColor))
		tab:ClearAllPoints()
		tab:SetPoint("CENTER", UIParent, "BOTTOMLEFT", db.TabX, db.TabY)

		do
			local backdrop = db.TabBackdrop
			local bg, edge = backdrop.bgFile, backdrop.edgeFile

			backdrop.bgFile, backdrop.edgeFile = MediaHandler:Fetch("background", bg), MediaHandler:Fetch("border", edge)
			tab:SetBackdrop(backdrop)

			backdrop.bgFile, backdrop.edgeFile = bg, edge
		end

		tab:SetBackdropColor(unpack(db.TabColor))
		tab:SetBackdropBorderColor(unpack(db.TabBorderColor))

	end
	function DockClass:ReclaimTab()
		local tab = self.Tab
		self.Tab = nil
		tab:ClearAllPoints()
		tab:Hide()
		tabs[#tabs + 1] = tab
	end
	function DockClass:AcquireTab()
		local tab
		if #tabs > 0 then
			tab = tabs[#tabs]
			tabs[#tabs] = nil
		else
			tab = NewTab()
		end

		tab:Show()
		self.Tab = tab
		tab.Dock = self

		self:UpdateTab()
	end
end

local function GetTableKey(db, ...)
	local num = select("#", ...)
	for i=num, 1, -1 do
		local key = select(i, ...)
		if db[key] ~= nil then
			local opt = db
			for i=i, num - 1 do
				local t = opt[select(i, ...)]
				if type(t) == "table" then
					opt = t
				else
					return opt, select(num - i, ...)
				end
			end
			return opt, select(num, ...)
		end
	end
end

function DockClass:SetOption(info, ...)
	local table, key = GetTableKey(self.db.profile, unpack(info))
	if type(table[key]) == "table" then --color table
		for i=1, select("#", ...) do
			table[key][i] = select(i, ...)
		end
	else
		table[key] = ...
	end
	
	self:RefreshSettings()
end

function DockClass:GetOption(info)
	local table, key = GetTableKey(self.db.profile, unpack(info))
	if type(table[key]) == "table" then --color table
		return unpack(table[key])
	end
	return table[key]
end

function DockClass:GetDisabledOption(info)
	local key = info[#info]
	if key == "ForcedWidth" then
		return not self.db.profile.ForceWidth
	elseif key == "ForcedHeight" then
		return not self.db.profile.ForceHeight
	elseif key == "AutoTab" or key == "TabOrientation" or key == "TabFont" or key == "TabFontColor" or 
		key == "TabAnimation" or key == "TabAnimationRate" or key == "TabColor" or key == "TabColor" or
		key == "TabHighlightColor" or key == "TabBorderColor" or key == "TabBorderHighlightColor" or key == "TabBackdrop"
		or key == "TabSize" or key == "TabFontSize" then

		return not self.db.profile.AllowTabMode
	elseif key == "AutoTabDelay" then
		return not (self.db.profile.AllowTabMode and self.db.profile.AutoTab)
	end
end