local module = Dock:NewModule("Chat Windows", "AceEvent-3.0")

local db, docked
local chatFrames = {
	[ChatFrame1] = true,
	[ChatFrame2] = true,
	[ChatFrame3] = true,
	[ChatFrame4] = true,
	[ChatFrame5] = true,
	[ChatFrame6] = true,
	[ChatFrame7] = true,
}
local function UpdateLanguageButton()
	if SELECTED_CHAT_FRAME.isDocked then
		ChatFrameMenuButton:SetParent(SELECTED_CHAT_FRAME)
	else
		ChatFrameMenuButton:SetParent(DEFAULT_CHAT_FRAME)
	end
	if ChatFrameMenuButton:IsShown() then
		ChatFrameMenuButton:Show()
	end
end
function module:OnInitialize()
	self.db:RegisterDefaults{
		profile = {
			ChatFrames = {
				["*"] = {
					Enabled = false,
					Style = "blizzard",
					Size = 16,
					Dock = "",
					XOffset = 0,
					YOffSet = 0,
				},
			},
			Docked = {},
		},
	}
	hooksecurefunc("FCF_SelectDockFrame", UpdateLanguageButton)
	
	UpdateLanguageButton()
end


function module:Dock_FrameDocked(event, frame, dock)
	if not chatFrames[frame] then
		return
	end
	if frame == DEFAULT_CHAT_FRAME then
		for i, f in ipairs(DOCKED_CHAT_FRAMES) do
			local tab = _G[f:GetName().."Tab"]
			tab:SetParent(frame)
			tab:Disable()
		end
	end
	_G[frame:GetName().."Tab"]:SetParent(frame)
	_G[frame:GetName().."Tab"]:Disable()
	UpdateLanguageButton()
end

function module:Dock_FrameReleased(event, frame, dock)
	if not chatFrames[frame] then
		return
	end
	if frame == DEFAULT_CHAT_FRAME then
		for i, f in ipairs(DOCKED_CHAT_FRAMES) do
			local tab = _G[f:GetName().."Tab"]
			tab:SetParent(UIParent)
			tab:Enable()
			if tab:IsShown() then
				tab:Show()
			end
		end
	end
	local tab = _G[frame:GetName().."Tab"]
	tab:SetParent(UIParent)
	tab:Enable()
	if tab:IsShown() then
		tab:Show()
	end
	UpdateLanguageButton()
end

function module:Dock_FrameRestored(event, frame)
	if docked[frame] and not frame.isDocked then
		FCF_DockFrame(frame, docked[frame], true)
		docked[frame] = nil
	end
end

function module:Dock_FrameSentToDock(event, frame, dock, button)
	if not chatFrames[frame] then
		return
	end
	if frame == DEFAULT_CHAT_FRAME then
		return
	end
	if frame.isDocked then
		for i, f in ipairs(DOCKED_CHAT_FRAMES) do
			if frame == f then
				docked[frame] = i
				break
			end
		end
	else
		docked[frame] = nil
	end
	FCF_UnDockFrame(frame)
	FCF_SetLocked(frame)
	_G[frame:GetName().."Tab"]:ClearAllPoints()
	FCF_SetTabPosition(frame, 0)
end

local UpdateButtons
do
	local function OnDragStart(self)
		self:SetMovable(true)
		self:SetClampedToScreen(true)
		local frame = self:GetFrame()
		local scale = frame:GetScale()
		local db = db.profile.ChatFrames[frame:GetName()]
		local leeway = db.Style == "blizzard" and db.Size * 2 or db.Size
		self:SetClampRectInsets(
			-frame:GetLeft() + leeway, 
			UIParent:GetRight() / scale - frame:GetRight() - leeway, 
			UIParent:GetTop() / scale - frame:GetTop() - leeway,
			-frame:GetBottom() + leeway
		)
		self:StartMoving()
	end

	local function OnDragStop(self)
		self:StopMovingOrSizing()
		self:SetMovable(false)
		self:SetClampedToScreen(false)
		self:SetClampRectInsets(0, 0, 0, 0)
		local frame = self:GetFrame()
		local bx, by = self:GetRight(), self:GetTop()
		local fx, fy = frame:GetRight(), frame:GetTop()
		local db = db.profile.ChatFrames[frame:GetName()]
		db.XOffset, db.YOffset = bx - fx, by - fy
		UpdateButtons()
	end

	function UpdateButtons()
		local db = db.profile.ChatFrames
		for i=1, 7 do
			local db = db["ChatFrame"..i]
			if db.Enabled then
				local frame = _G["ChatFrame"..i]
				local button = Dock:GetDockButton(frame) or Dock:NewDockButton(frame)

				button:ClearAllPoints()
				button:SetPoint("TOPRIGHT", frame, "TOPRIGHT", db.XOffset, db.YOffset)

				if db.Style == "blizzard" then
					button:SetHeight(db.Size * 2)
					button:SetWidth(db.Size * 2)
				else
					button:SetHeight(db.Size)
					button:SetWidth(db.Size)
				end

				button:SetStyle(db.Style)
				button:SetDock(db.Dock)
				button:RegisterForDrag("LeftButton")
				button:SetScript("OnDragStart", OnDragStart)
				button:SetScript("OnDragStop", OnDragStop)
			else
				local button = Dock:GetDockButton(_G["ChatFrame"..i])
				if button then
					button:RegisterForDrag()
					button:SetScript("OnDragStart", nil)
					button:SetScript("OnDragStop", nil)
					button:Release()
				end
			end
		end
	end
end

function module:OnEnable()
	db = self.db
	docked = db.profile.Docked
	UpdateButtons()
	self:RegisterMessage("Dock_FrameRestored")
	self:RegisterMessage("Dock_FrameSentToDock")
	self:RegisterMessage("Dock_FrameDocked")
	self:RegisterMessage("Dock_FrameReleased")
end

function module:OnDisable()
	for i=1, 7 do
		local button = Dock:GetDockButton(_G["ChatFrame"..i])
		if button then
			button:Release()
		end
	end
end

function module:GetOptionsTable()
	module.GetOptionsTable = nil
	local opts = {
		type = "group",
		name = "Chat Windows",
		childGroups = "tree",
		get = function(info)
			local frame, key = info[#info - 1], info[#info]
			local value = db.profile.ChatFrames[frame][key]
			if key == "Dock" and not Dock:HasDockFrame(value) then
				return ""
			end
			return value
		end,
		set = function(info, value)
			db.profile.ChatFrames[info[#info - 1]][info[#info]] = value
			UpdateButtons()
		end,
		args = {},
	}
	local function GetDocks()
		local docks = {[""] = "None"}
		for name in Dock:IterateDockFrames() do
			docks[name] = name
		end
		return docks
	end 
	local chatOpts = {
		type = "group",
		name = function(info)
			return info[#info]:match("%d") .. ". " .. _G[info[#info].."Tab"]:GetText()
		end,
		disabled = function()
			return not self:IsEnabled()
		end,
		args = {
			Enabled = {
				type = "toggle",
				name = "Add button",
				desc = "Adds a send to dock button to this chat frame.",
				order = 10,
			},
			Dock = {
				type = "select",
				name = "Dock",
				desc = "Sets the dock that the chat windows should be sent to.",
				values = GetDocks,
				order = 20,
			},
			Style = {
				type = "select",
				name = "Style",
				desc = "Sets the button style used.",
				values = {blizzard = "Blizzard", modern = "Modern"},
				order = 30,
			},
			Size = {
				type = "range",
				name = "Size",
				desc = "Sets the size of the button.",
				min = 1,
				max = 100,
				bigStep = 1,
				order = 40,
			},
		},
	}
	for i=1, 7 do
		opts.args["ChatFrame"..i] = chatOpts
	end
	return opts
end