local pairs = pairs
local ipairs = ipairs
local type = type

setfenv(1, Dock.FunctionEnviroment)

local Core = Core

ConfigLoaded = true


local docks = {}
local bindings = {}

local function SetOption(info, value)
	Core.db.profile[info[#info]] = value
end

local function GetOption(info)
	return Core.db.profile[info[#info]]
end

function ValidateDockName(info, name)
	if name == "" or not name:match("%S") then
		return "A dock's name cannot be empty"
	elseif Core:HasDockFrame(name) then
		return "A dock with that name already exists"
	end
	return true
end

local function GetDocks()
	local docks = {[""] = "None"}
	for name in Core:IterateDockFrames() do
		docks[name] = name
	end
	return docks
end

do
	local function GetBindingOption(info)
		local name = Core.db.profile[info[#info]]
		if Core:HasDockFrame(name) then
			return name
		end
		return ""
	end
	local function GetBinding(info)
		local key = _G.GetBindingKey("Send to dock " .. info[#info])
		return key ~= "" and key or "Unbound"
	end
	local function SetBinding(info, key)
		local cmd = "Send to dock " .. info[#info]
		local old = _G.GetBindingKey(cmd)
		if old then
			_G.SetBinding(old, nil)
		end
		_G.SetBinding(key, cmd)
		_G.SaveBindings(_G.GetCurrentBindingSet())
	end
	bindings.Descrip = {
		type = "description",
		name = "Keybinds let you send a frame to a dock by mousing over the frame and pressing the keybind.",
		order = .5,
	}
	for i=1, 9 do
		bindings[i..""] = {
			type = "keybinding",
			name = ("Binding %d"):format(i),
			desc = ("Sets the key used in binding %d."):format(i),
			set = SetBinding,
			get = GetBinding,
			order = i,
		}
		bindings["DockBinding"..i] = {
			type = "select",
			name = ("Binding dock %d"):format(i),
			desc = ("Sets binding %d to use the dock selected here."):format(i),
			set = SetOption,
			get = GetBindingOption,
			values = GetDocks,
			order = i + .1,
		}
		bindings["DockBindingInstant"..i] = {
			type = "toggle",
			name = ("Instant send %d"):format(i),
			desc = ("Sets binding %d to send the frame to the dock instantly instead of using an animation."):format(i),
			set = SetOption,
			get = GetOption,
			order = i + .2,
		}
		bindings["DockBindingHeader"..i] = {
			type = "header",
			name = "",
			order = i - .1,
		}
	end
end

local modules = {}


local currentCopyProfile, currentDeleteProfile
local options = {
	type = "group",
	name = "Dock",
	handler = Core,
	childGroups = "tree",
	args = {
		NewDock = {
			type = "input",
			name = "New dock",
			desc = "Creates a new dock. You may have an unlimited amount of docks, all with their own settings.\n\nType the name of the dock you wish to create and press enter or OK.",
			validate = ValidateDockName,
			get = false,
			set = function(info, name) Core:NewDock(name) end,
			order = .1,
		},
		Enabled = {
			type = "toggle",
			name = "Enabled",
			desc = "Controls the enabled state of Dock. Disabling Dock will hide all docks and places the addon into standby.",
			get = function() return Core:IsEnabled() end,
			set = function()
				if Core:IsEnabled() then
					Core:Disable()
					Core.db.global.Enabled = false
				else
					Core:Enable()
					Core.db.global.Enabled = true
				end
			end,
			order = .2,
		},
		ReportFailure = {
			type = "toggle",
			name = "Report failures",
			desc = "Reports the reason why a frame failed to dock in the chat window.",
			get = GetOption,
			set = SetOption,
			order = .3,
		},
		Modules = {
			type = "group",
			name = "Modules",
			childGroups = "tree",
			order = 10,
			args = modules,
		},
		Docks = {
			type = "group",
			name = "Docks",
			childGroups = "tree",
			order = 20,
			args = docks,
		},
		Bindings = {
			type = "group",
			name = "Bindings",
			childGroups = "select",
			order = 40,
			args = bindings,
		},
		Profiles = {
			type = "group",
			name = "Profiles",
			childGroups = "select",
			order = 50,
			args = {
				Description = {
					type = "description",
					name = "Profiles contain settings for all docks and modules. By default Dock stores each character's settings separately, by selecting another character's profile the two (or more) characters will share settings.",
					order = 1,
				},
				GeneralBreak = {type = "header", order = 5, name = "General",},
				CurrentProfile = {
					type = "select",
					name = "Current profile",
					desc = "Sets the current profile to the one selected here.",
					values = function()
						local profiles = db:GetProfiles()
						for i, v in ipairs(profiles) do
							profiles[i] = nil
							profiles[v] = v
						end
						return profiles
					end,
					get = function() return db:GetCurrentProfile() end,
					set = function(info, name) db:SetProfile(name) end,
					width = "full",
					order = 10,
				},
				NewProfile = {
					type = "input",
					name = "New profile",
					desc = "Creates a new profile.\n\nType the name of the profile you wish to create and press enter or OK.",
					validate = function(info, name)
						if name == "" or not name:match("%S") then
							return "A profile name cannot be empty"
						end
						return true
					end,
					get = false,
					set = function(info, name) db:SetProfile(name) end,
					width = "full",
					order = 20,
				},
				CopyBreak = {type = "header", order = 30, name = "Copying",},
				SelectCopyProfile = {
					type = "select",
					name = "Select profile to copy from",
					desc = "Select a profile to copy from.",
					values = function()
						local profiles = db:GetProfiles()
						local current = db:GetCurrentProfile()
						if #profiles > 1 then
							for i, v in ipairs(profiles) do
								profiles[i] = nil
								if current ~= v then
									profiles[v] = v
								end
							end
							profiles["None"] = "None"
							return profiles
						else
							return {["None"] = "None",}
						end
					end,
					get = function()
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentCopyProfile then
								return name == db:GetCurrentProfile() and "None" or name
							end
						end
						return "None"
					end,
					set = function(info, name)
						currentCopyProfile = name
					end,
					width = "full",
					order = 40,
				},
				CopyProfile = {
					type = "execute",
					name = function()
						local current = "None"
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentCopyProfile then
								if name ~= db:GetCurrentProfile() then
									current = currentCopyProfile
								end
								break
							end
						end
						return ("Copy profile from %q to %q"):format(current, db:GetCurrentProfile())
					end,
					desc = "The profile selected above is copied into the current profile.",
					disabled = function()
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentCopyProfile then
								return name == db:GetCurrentProfile()
							end
						end
						return true
					end,
					order = 50,
					width = "full",
					func = function() db:CopyProfile(currentCopyProfile) end,
				},
				DeleteBreak = {type = "header", order = 60, name = "Removal",},
				SelectDeleteProfile = {
					type = "select",
					name = "Select a profile to remove",
					desc = "Select a profile to remove.",
					values = function()
						local profiles = db:GetProfiles()
						local current = db:GetCurrentProfile()
						if #profiles > 1 then
							for i, v in ipairs(profiles) do
								profiles[i] = nil
								if current ~= v then
									profiles[v] = v
								end
							end
							profiles["None"] = "None"
							return profiles
						else
							return {["None"] = "None",}
						end
					end,
					get = function()
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentDeleteProfile then
								return name == db:GetCurrentProfile() and "None" or name
							end
						end
						return "None"
					end,
					set = function(info, name)
						currentDeleteProfile = name
					end,
					width = "full",
					order = 70,
				},
				DeleteProfile = {
					type = "execute",
					name = function()
						local current = "None"
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentDeleteProfile then
								if name ~= db:GetCurrentProfile() then
									current = name
								end
								break
							end
						end
						return ("Delete profile %q"):format(current)
					end,
					desc = "The profile selected above removed. This will permanently remove it, along with all dock and module settings associated with it.",
					disabled = function()
						for i, name in ipairs(db:GetProfiles()) do
							if name == currentDeleteProfile then
								return name == db:GetCurrentProfile()
							end
						end
						return true
					end,
					order = 80,
					width = "full",
					func = function() db:DeleteProfile(currentDeleteProfile) end,
					confirm = true,
					confirmText = "This will permanently remove the profile and its settings, are you sure?",
				},
				MaintenanceBreak = {type = "header", order = 90, name = "Maintenance",},
				ResetProfile = {
					type = "execute",
					name = "Reset profile",
					desc = "Resets the current profile to its default values. This will reset all modules to their default values and removes all docks for this profile.",
					order = 100,
					func = function() db:ResetProfile() end,
				},
				ResetDB = {
					type = "execute",
					name = "Reset database",
					desc = "Permanently resets the ENTIRE database to default values. This will reset all modules to their default values, removes all docks, and remove every profile.",
					order = 110,
					func = function() db:ResetDB() end,
				},
			},
		},
	},
		
}


local dockOptions

local function GetHandler(info)
	local opts = options.args
	local handler = Core
	for i=1, #info - 1 do
		opts = opts[info[i]]
		handler = opts.handler or handler
		opts = opts.args
	end
	return handler
end


local UpdateModules
do
	ToggleActive = {
		type = "toggle",
		name = "Enabled",
		desc = "Sets the enabled state of this module.",
		get = function(info)
			return GetHandler(info):IsEnabled()
		end,
		set = function(info)
			local module = GetHandler(info)
			if module:IsEnabled() then
				module:Disable()
				Core.db.profile.ModuleStates[module.name] = false
			else
				module:Enable()
				Core.db.profile.ModuleStates[module.name] = true
			end
		end,
		disabled = false,
		order = -.001,
	}
	function UpdateModules()
		for name, module in Core:IterateModules() do
			local strippedName = name:gsub("[%c \127]", "_")
			if not modules[strippedName] and type(module.GetOptionsTable) == "function" then
				local opts = module:GetOptionsTable()
				if opts then
					if not opts.handler then
						opts.handler = module
					end
					local args = opts.args
					if args then
						if args.ToggleActive == nil then
							args.ToggleActive = ToggleActive
						end
					end
					modules[strippedName] = opts
				end
				
			end
		end
		options.args.Modules.hidden = not _G.next(modules) and true or nil
	end
end

function Core:RefreshAndReturnConfig()
	for name, opts in pairs(docks) do --clean out removed docks
		if not Core:HasDockFrame(opts.name) then
			docks[name] = nil
		end
	end

	for name, dock in self:IterateDockFrames() do
		local strippedName = name:gsub("[%c \127]", "_")
		if not docks[strippedName] then
			docks[strippedName] = {
				type = "group",
				name = name,
				handler = dock,
				set = "SetOption",
				get = "GetOption",
				disabled = "GetDisabledOption",
				args = dockOptions,
				childGroups = "tab",
			}
		else
			docks[strippedName].handler = dock --update it incase a profile switch that uses the same dock name
		end
		
	end
	UpdateModules()

	return options
end

local function BuildMediaValues(mediaType)
	return function()
		local dict = {}
		for i, name in ipairs(MediaHandler:List(mediaType)) do
			dict[name] = name
		end
		return dict
	end
end

local dockedFrameClick = {
	nothing = "Do nothing",
	release = "Restore frame",
	lockZoom = "Lock frame's zoom",
	releaseAll = "Restore ALL docked frames",
	tabMode = "Change dock to tab mode",
}
local dockClick = {
	nothing = "Do nothing",
	config = "Open config",
	releaseAll = "Restore ALL docked frames",
	tabMode = "Change dock to tab mode",
}

local copySelectedDock
dockOptions = {
	General = {
		type = "group",
		name = "General",
		order = 10,
		args = {
			RemoveDock = {
				type = "execute",
				name = "Remove dock",
				desc = "Permanently removes this dock and its settings.",
				order = 10,
				func = "GUIRemove",
				confirm = true,
				confirmText = "This will permanently remove the dock and its settings, are you sure?",
			},
			Locked = {
				type = "toggle",
				name = "Locked",
				desc = "Locks the dock in place keeping it from being dragged or resized.",
				order = 20,
			},
			Rename = {
				type = "input",
				name = "Rename dock",
				desc = "Renames this dock to something else, keeping all settings intact.",
				order = 30,
				get = function(info)
					return docks[info[#info - 2]].name
				end,
				set = function(info, name)
					Core:RenameDockFrame(docks[info[#info - 2]].name, name)
				end,
				validate = ValidateDockName,
			},
			Clone = {
				type = "input",
				name = "Clone dock",
				desc = "Clones this dock with the name specified, copying all settings (except position and docked frames) to the new dock.",
				order = 40,
				get = false,
				set = function(info, name)
					Core:CopyDockFrame(docks[info[#info - 2]].name, name)
				end,
				validate = ValidateDockName,
			},
			Break = {type ="header", order = 50, name = "",},
			CopyDesc = {
				type = "description",
				order = 60,
				name = "Select a dock to copy this dock's settings to then press the 'Copy to selected' button. Copying settings will transfer all settings over except the position and saved docked frames.",
			},
			CopySelect = {
				type = "select",
				name = "Select dock",
				desc = "Selects a dock to copy settings to.",
				order = 70,
				get = function(info)
					return Core:HasDockFrame(copySelectedDock) and copySelectedDock or "" 
				end,
				set = function(info, name)
					copySelectedDock = name
				end,
				values = GetDocks,
			},
			CopyDock = {
				type = "execute",
				name = "Copy to selected",
				desc = "Copies settings from this dock to the dock selected.",
				order = 80,
				func = function(info)
					Core:CopyDockFrame(docks[info[#info - 2]].name, copySelectedDock)
				end,
				disabled = function()
					return not Core:HasDockFrame(copySelectedDock)
				end,
			},
		},
	},
	Backdrop = {
		type = "group",
		name = "Backdrop",
		order = 15,
		args = {
			bgFile = {
				type = "select",
				name = "Background",
				desc = "The texture used for the dock's backdrop.",
				values = BuildMediaValues("background"),
				order = 10,
			},
			tile = {
				type = "toggle",
				name = "Tile",
				desc = "Whether or not the backdrop should be tiled instead of being stretched.",
				order = 30,
			},
			tileSize = {
				type = "range",
				name = "Tile size",
				desc = "The tile size of the backdrop if tiled is selected.",
				min = 0,
				max = 50,
				bigStep = 1,
				order = 40,
			},
			edgeFile = {
				type = "select",
				name = "Border",
				desc = "The texture used for the dock's border.",
				values = BuildMediaValues("border"),
				order = 50,
			},
			edgeSize = {
				type = "range",
				name = "Border size",
				desc = "The size of the border.",
				min = 0,
				max = 50,
				bigStep = 1,
				order = 60,
			},
			insets = {
				type = "group",
				name = "Insets",
				order = 70,
				inline = true,
				args = {
					left = {
						type = "range",
						name = "Left inset",
						desc = "Controls the space between the edge of the backdrop texture and the edge of the dock.",
						min = 0,
						max = 50,
						bigStep = 1,
						order = 10,
					},
					right = {
						type = "range",
						name = "Right inset",
						desc = "Controls the space between the edge of the backdrop texture and the edge of the dock.",
						min = 0,
						max = 50,
						bigStep = 1,
						order = 20,
					},
					bottom = {
						type = "range",
						name = "Bottom inset",
						desc = "Controls the space between the edge of the backdrop texture and the edge of the dock.",
						min = 0,
						max = 50,
						bigStep = 1,
						order = 30,
					},
					top = {
						type = "range",
						name = "Top inset",
						desc = "Controls the space between the edge of the backdrop texture and the edge of the dock.",
						min = 0,
						max = 50,
						bigStep = 1,
						order = 40,
					},
				},
			},
		},
	},
	Anchoring = {
		type = "group",
		name = "Anchoring",
		order = 20,
		args = {
			HorizontalAnchor = {
				type = "select",
				name = "Horizontal anchor",
				desc = "Determines the side of the dock that frames anchor to, and thus the point they zoom from.",
				order = 10,
				values = {
					LEFT = "Left",
					RIGHT = "Right",
				},
				style = "radio",
			},
			VerticalAnchor = {
				type = "select",
				name = "Vertical anchor",
				desc = "Determines the side of the dock that frames anchor to, and thus the point they zoom from.",
				order = 20,
				values = {
					TOP = "Top",
					BOTTOM = "Bottom",
				},
				style = "radio",
			},
			Orientation = {
				type = "select",
				name = "Orientation",
				desc = "Determines the orientation of the dock.",
				order = 30,
				values = {
					Horizontal = "Horizontal",
					Vertical = "Vertical",
				},
				style = "radio",
			},
			Border = {
				type = "range",
				name = "Frame spacing",
				desc = "Determines the amount of distance between the edge of the dock and between docked frames.",
				min = 0,
				max = 50,
				bigStep = 1,
				order = 40,
			},
		},
	},
	Coloring = {
		type = "group",
		name = "Coloring",
		order = 30,
		args = {
			BackdropColor = {
				type = "color",
				name = "Backdrop color",
				desc = "The color of the dock's backdrop.",
				hasAlpha = true,
				order = 10,
			},
			BackdropHighlightColor = {
				type = "color",
				name = "Backdrop highlight color",
				desc = "The color of the dock's backdrop when the mouse is inside the dock.",
				hasAlpha = true,
				order = 20,
			},
			BackdropBorderColor = {
				type = "color",
				name = "Border color",
				desc = "The color of the dock's border.",
				hasAlpha = true,
				order = 30,
			},
			BackdropBorderHighlightColor = {
				type = "color",
				name = "Border highlight color",
				desc = "The color of the dock's border when the mouse is inside the dock.",
				hasAlpha = true,
				order = 40,
			},
			ArrowColor = {
				type = "color",
				name = "Dock arrow color",
				desc = "The color of the docking arrow.",
				hasAlpha = true,
				order = 50,
			},
			OverlayColor = {
				type = "color",
				name = "Overlay color",
				desc = "The color of the frame overlay on docked frames.",
				hasAlpha = true,
				order = 60,
			},
			OverlayHighlightColor = {
				type = "color",
				name = "Overlay highlight color",
				desc = "The color of the frame overlay on docked frames when the mouse is over it.",
				hasAlpha = true,
				order = 70,
			},
			TabColor = {
				type = "color",
				name = "Tab backdrop color",
				desc = "The color of the tab's backdrop.",
				hasAlpha = true,
				order = 80,
			},
			TabHighlightColor = {
				type = "color",
				name = "Tab backdrop highlight color",
				desc = "The color of the tab's backdrop when the mouse is inside the tab.",
				hasAlpha = true,
				order = 90,
			},
			TabBorderColor = {
				type = "color",
				name = "Tab border color",
				desc = "The color of the tabs's border.",
				hasAlpha = true,
				order = 100,
			},
			TabBorderHighlightColor = {
				type = "color",
				name = "Tab border highlight color",
				desc = "The color of the tab's border when the mouse is inside the tab.",
				hasAlpha = true,
				order = 110,
			},
			TabFontColor = {
				type = "color",
				name = "Tab font color",
				desc = "The color of the tab's font.",
				order = 120,
			},
		},
	},
	Clicks = {
		type = "group",
		name = "Clicks",
		order = 40,
		args = {
			OverlayLeftButton = {
				type = "select",
				name = "Frame Left click",
				desc = "The action to take when a docked frame is clicked on.",
				values = dockedFrameClick,
				order = 10,
			},
			OverlayMiddleButton = {
				type = "select",
				name = "Frame Middle click",
				desc = "The action to take when a docked frame is clicked on.",
				values = dockedFrameClick,
				order = 20,
			},
			OverlayRightButton = {
				type = "select",
				name = "Frame Right click",
				desc = "The action to take when a docked frame is clicked on.",
				values = dockedFrameClick,
				order = 30,
			},
			Break = {type="header", name = "", order = 40},
			DockMiddleButton = {
				type = "select",
				name = "Dock Middle click",
				desc = "The action to take when the dock is clicked on.",
				values = dockClick,
				order = 50,
			},
			DockRightButton = {
				type = "select",
				name = "Dock Right click",
				desc = "The action to take when the dock is clicked on.",
				values = dockClick,
				order = 60,
			},
		},
	},
	Tab = {
		type = "group",
		name = "Tab",
		desc = "Recently added and not thoroughly tested, use at your own risk!",
		order = 45,
		args = {
			Disclaimer = {
				type = "description",
				name = "THIS FEATURE HAS NOT BEEN THOROUGHLY TESTED, USE AT YOUR OWN RISK.Please report bugs if you do use this feature.",
				order = 1,
			},
			Desc = {
				type = "description",
				name = "Tab mode lets you collapse docks even further into tabs. Docks that are in tab mode can still have frames dragged to them and keybinds will still function normally. Simply clicking the tab will restore the dock to its normal function.",
				order = 10,
			},
			AllowTabMode = {
				type = "toggle",
				name = "Enable tab mode",
				desc = "Allows tab mode. Set 'change dock to tab mode' under the 'Clicks' option tab.",
				order = 20,
			},
			AutoTab = {
				type = "toggle",
				name = "Auto tab",
				desc = "Automatically changes the dock into tab mode if the mouse isn't near the dock.",
				order = 30,
			},
			AutoTabDelay = {
				type = "range",
				name = "Auto tab delay",
				desc = "Determines the time to wait (in seconds) after the mouse leaves before automatically changing the dock into tab mode.",
				min = 0,
				max = 20,
				bigStep = 1,
				order = 40,
			},
			TabOrientation = {
				type = "select",
				name = "Tab orientation",
				desc = "Determines the orientation of the tab.",
				order = 50,
				values = {
					Horizontal = "Horizontal",
					Vertical = "Vertical",
				},
				style = "radio",
			},
			TabSize = {
				type = "range",
				name = "Tab size",
				desc = "The size of the tab adjusted by the size of the font.",
				min = 0,
				max = 100,
				bigStep = 1,
				order = 60,
			},
			TabFont = {
				type = "select",
				name = "Font",
				desc = "The font used on the tab's label.",
				values = BuildMediaValues("font"),
				order = 70,
			},
			TabFontSize = {
				type = "range",
				name = "Font size",
				desc = "Determines the size of the font used on the tab's label.",
				min = 8,
				max = 26,
				bigStep = 1,
				order = 80,
			},

			
		},
	},
	Animations = {
		type = "group",
		name = "Animations",
		order = 50,
		args = {
			Zooming = {
				type = "group",
				name = "Zooming",
				order = 10,
				inline = true,
				args = {
					MouseOverSize = {
						type = "range",
						name = "Mouseover zoom",
						desc = "How much to zoom in when a frame is moused over.",
						min = .01,
						max = .99,
						bigStep = .01,
						isPercent = true,
						order = 10,
					},
					MouseDetect = {
						type = "range",
						name = "Zoom detect",
						desc = "How close your mouse needs to be to start zooming in on a frame. Smaller values will cause frames to zoom in quickly once the mouse is close to the dock, while larger values will start zooming in further from the dock but less quickly.",
						min = 50,
						max = 300,
						bigStep = 1,
						order = 20,
					},
				},
			},
			Undocking = {
				type = "group",
				name = "Undocking",
				order = 20,
				inline = true,
				args = {
					Animation = {
						type = "select",
						name = "Undock animation style",
						desc = "The animation style when docked frames are clicked on.",
						values = Core:GetAnimationList(),
						order = 10,
						width = "double",
					},
					AnimationRate = {
						type = "range",
						name = "Undock animation speed",
						desc = "How quickly the animation should be drawn when a frame is undocked. 1% is very very slow, and 100% is near instant.",
						min = .01,
						max = 1,
						bigStep = .01,
						isPercent = true,
						width = "double",
						order = 20,
					},
				},
			},
			Docking = {
				type = "group",
				name = "Docking",
				order = 30,
				inline = true,
				args = {
					DockAnimation = {
						type = "select",
						name = "Dock animation style",
						desc = "The animation style when frames are sent to the dock.",
						values = Core:GetAnimationList(),
						order = 10,
						width = "double",
					},
					DockAnimationRate = {
						type = "range",
						name = "Dock animation speed",
						desc = "How quickly the animation should be drawn when frames are sent to the dock. 1% is very very slow, and 100% is near instant.",
						min = .01,
						max = 1,
						bigStep = .01,
						isPercent = true,
						width = "double",
						order = 20,
					},
				},
			},
			Tabs = {
				type = "group",
				name = "Tabs",
				order = 40,
				inline = true,
				args = {
					TabAnimation = {
						type = "select",
						name = "Tab animation style",
						desc = "The animation style when docks are sent to tab mode and vice versa.",
						values = Core:GetAnimationList(),
						order = 10,
						width = "double",
					},
					TabAnimationRate = {
						type = "range",
						name = "Tab animation speed",
						desc = "How quickly the animation should be drawn when docks are sent to tab mode or vice versa. 1% is very very slow, and 100% is near instant.",
						min = .01,
						max = 1,
						bigStep = .01,
						isPercent = true,
						width = "double",
						order = 20,
					},
				},
			},
		},
	},
	Advanced = {
		type = "group",
		name = "Advanced",
		order = 70,
		args = {
			AdvancedDescrip = {
				type = "description",
				order = 5,
				name = "Advanced options are usually best left at default values unless you are looking for very specific behavior.",
			},
			Break1 = {type ="header", order = 9, name = "",},
			ValidateFrameDelay = {
				type = "range",
				name = "Frame validation delay",
				desc = "How long to wait before a frame saved in this dock should be checked for re-docking after it's created.\n\nIncreasing this value can sometimes help re-dock stubborn frames from previous sessions.",
				min = 0,
				max = 10,
				bigStep = 1,
				order = 10,
			},
			Break2 = {type ="header", order = 20, name = "",},
			ForceDescrip = {
				type = "description", 
				order = 25, 
				name = "Enabling these options will force the width or height of docked frames to match the sliders. If both options are checked then whichever value is larger is used, which will depend on the frame since each frame has different dimensions.",},
			ForceWidth = {
				type = "toggle",
				name = "Force width",
				desc = "Forces docked frames to be the width defined by the 'Forced width' slider.",
				order = 30,
			},
			ForcedWidth = {
				type = "range",
				name = "Forced width",
				desc = "Sets the width used when calculating the scale of docked frames.",
				min = 1,
				max = 1000,
				bigStep = 1,
				order = 40,
			},
			ForceHeight = {
				type = "toggle",
				name = "Force height",
				desc = "Forces docked frames to be the height defined by the 'Forced height' slider.",
				order = 50,
			},
			ForcedHeight = {
				type = "range",
				name = "Forced height",
				desc = "Sets the height used when calculating the scale of docked frames.",
				min = 1,
				max = 1000,
				bigStep = 1,
				order = 60,
			},
		},
	},
}